package ru.dbotthepony.kstarbound.world

import ru.dbotthepony.kbox2d.api.BodyDef
import ru.dbotthepony.kbox2d.api.FixtureDef
import ru.dbotthepony.kbox2d.collision.shapes.PolygonShape
import ru.dbotthepony.kbox2d.dynamics.B2Fixture
import ru.dbotthepony.kstarbound.defs.tile.TileDefinition
import ru.dbotthepony.kstarbound.world.entities.Entity
import ru.dbotthepony.kstarbound.world.phys.RectTileFlooderDepthFirst
import ru.dbotthepony.kstarbound.world.phys.RectTileFlooderSizeFirst
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import java.util.*
import kotlin.collections.ArrayList
import kotlin.collections.HashSet

/**
 * Представляет из себя класс, который содержит состояние тайла на заданной позиции
 */
data class ChunkTile(val chunk: Chunk<*, *>.TileLayer, val def: TileDefinition) {
	var color = 0
		set(value) {
			field = value
			chunk.incChangeset()
		}

	var forceVariant = -1
		set(value) {
			field = value
			chunk.incChangeset()
		}
}

private fun ccwSortScore(point: Vector2d, axis: Vector2d): Double {
	if (point.x > 0.0) {
		return point.dot(axis)
	} else {
		return -2 - point.dot(axis)
	}
}

/**
 * Чанк мира
 *
 * Хранит в себе тайлы и ентити внутри себя
 *
 * Считается, что один тайл имеет форму квадрата и сторона квадрата примерно равна полуметру,
 * что будет называться Starbound Unit
 *
 * Весь игровой мир будет измеряться в Starbound Unit'ах
 */
abstract class Chunk<WorldType : World<WorldType, This>, This : Chunk<WorldType, This>>(val world: WorldType, val pos: ChunkPos) {
	/**
	 * Возвращает счётчик изменений чанка
	 */
	var changeset = 0
		private set

	fun incChangeset() {
		changeset++
	}

	val left get() = pos.left
	val right get() = pos.right
	val top get() = pos.top
	val bottom get() = pos.bottom
	val topLeft get() = pos.topLeft
	val topRight get() = pos.topRight
	val bottomLeft get() = pos.bottomLeft
	val bottomRight get() = pos.bottomRight

	val aabb = aabbBase + Vector2d(pos.x * CHUNK_SIZE.toDouble(), pos.y * CHUNK_SIZE.toDouble())

	var isPhysicsDirty = false

	fun markPhysicsDirty() {
		if (isPhysicsDirty)
			return

		isPhysicsDirty = true
		world.dirtyPhysicsChunks.add(this as This)
	}

	fun bakeCollisions() {
		foreground.bakeCollisions()
	}

	inner class TileLayer : IMutableTileChunk {
		/**
		 * Возвращает счётчик изменений этого слоя
		 */
		var changeset = 0
			private set

		fun incChangeset() {
			changeset++
			this@Chunk.changeset++
		}

		private val collisionCache = ArrayList<AABB>()
		private val collisionCacheView = Collections.unmodifiableCollection(collisionCache)
		private var collisionChangeset = -1

		private val body = world.physics.createBody(BodyDef(
			position = pos.firstBlock.toDoubleVector(),
			userData = this
		))

		private val collisionChains = ArrayList<B2Fixture>()

		fun bakeCollisions() {
			if (collisionChangeset == changeset)
				return

			collisionChangeset = changeset
			collisionCache.clear()

			for (box in collisionChains) {
				body.destroyFixture(box)
			}

			collisionChains.clear()

			val xyAdd = Vector2d(pos.x * CHUNK_SIZEd, pos.y * CHUNK_SIZEd)
			val seen = BooleanArray(CHUNK_SIZE * CHUNK_SIZE)

			for (y in 0 .. CHUNK_SIZE_FF) {
				for (x in 0..CHUNK_SIZE_FF) {
					if (!seen[x or (y shl CHUNK_SHIFT)] && tiles[x or (y shl CHUNK_SHIFT)] != null) {
						val depthFirst = RectTileFlooderDepthFirst(tiles, seen, x, y)
						val sizeFirst = RectTileFlooderSizeFirst(tiles, seen, x, y)
						val xSpanDepth = depthFirst.maxs.x - depthFirst.mins.x
						val ySpanDepth = depthFirst.maxs.y - depthFirst.mins.y
						val xSpanSize = sizeFirst.maxs.x - sizeFirst.mins.x
						val ySpanSize = sizeFirst.maxs.y - sizeFirst.mins.y

						val aabb: AABB

						if (xSpanDepth * ySpanDepth > xSpanSize * ySpanSize) {
							depthFirst.markSeen()
							aabb = AABB(depthFirst.mins.toDoubleVector(), depthFirst.maxs.toDoubleVector() + Vector2d.POSITIVE_XY)
						} else {
							sizeFirst.markSeen()
							aabb = AABB(sizeFirst.mins.toDoubleVector(), sizeFirst.maxs.toDoubleVector() + Vector2d.POSITIVE_XY)
						}

						collisionChains.add(body.createFixture(FixtureDef(
							shape = PolygonShape().also { it.setAsBox(aabb.width / 2.0, aabb.height / 2.0, aabb.centre, 0.0) },
							friction = 0.4,
						)))

						collisionCache.add(aabb + xyAdd)
					}
				}
			}

			/*
			val seen = BooleanArray(CHUNK_SIZE * CHUNK_SIZE)

			for (y in 0 .. CHUNK_SIZE_FF) {
				for (x in 0 .. CHUNK_SIZE_FF) {
					if (!seen[x or (y shl CHUNK_SHIFT)] && tiles[x or (y shl CHUNK_SHIFT)] != null) {
						val exposed = TileFlooder(tiles, seen, x, y).exposed

						if (exposed.size > 0) {
							val vertices = ArrayList<Vector2d>()
							var centroid = Vector2d.ZERO

							for (pos in exposed) {
								centroid += pos.pos
							}

							centroid /= exposed.size.toDouble()

							for (pos in exposed) {
								if (pos.left) {
									vertices.addU(pos.pos + Vector2d(-0.5, -0.5))
									vertices.addU(pos.pos + Vector2d(-0.5, 0.5))
								}

								if (pos.up) {
									vertices.addU(pos.pos + Vector2d(-0.5, 0.5))
									vertices.addU(pos.pos + Vector2d(0.5, 0.5))
								}

								if (pos.right) {
									vertices.addU(pos.pos + Vector2d(0.5, 0.5))
									vertices.addU(pos.pos + Vector2d(0.5, -0.5))
								}

								if (pos.down) {
									vertices.addU(pos.pos + Vector2d(0.5, -0.5))
									vertices.addU(pos.pos + Vector2d(-0.5, -0.5))
								}
							}

							for (i in vertices.indices) {
								vertices[i] -= centroid
							}

							vertices.sortWith { a, b ->
								val scoreA = ccwSortScore(a.normalized, Vector2d.POSITIVE_Y)
								val scoreB = ccwSortScore(b.normalized, Vector2d.POSITIVE_Y)

								if (scoreA == scoreB) {
									return@sortWith 0
								} else if (scoreA > scoreB) {
									return@sortWith 1
								} else {
									return@sortWith -1
								}
							}

							for (i in vertices.indices) {
								vertices[i] += centroid
							}

							val deflated = ArrayList<Vector2d>()
							var edge = vertices[0]
							deflated.add(edge)

							for (i in 1 until vertices.size) {
								if (vertices[i].x == edge.x) {
									// expand along X axis
									continue
								} else if (vertices[i].y == edge.y) {
									// expand along Y axis
									continue
								}

								// completely different direction

								if (deflated.indexOf(vertices[i - 1]) == -1)
									deflated.add(vertices[i - 1])

								if (deflated.indexOf(vertices[i]) == -1)
									deflated.add(vertices[i])

								edge = vertices[i]
							}

							collisionChains.add(body.createFixture(FixtureDef(
								shape = ChainShape().also { it.createLoop(vertices) },
								friction = 0.4,
							)))
						}
					}
				}
			}*/
		}

		/**
		 * Возвращает список AABB тайлов этого слоя
		 *
		 * Данный список напрямую указывает на внутреннее состояние и будет изменён при перестройке
		 * коллизии чанка, поэтому если необходим стабильный список, его необходимо скопировать
		 */
		fun collisionLayers(): Collection<AABB> {
			if (collisionChangeset != changeset) {
				bakeCollisions()
			}

			return collisionCacheView
		}

		override val pos: ChunkPos
			get() = this@Chunk.pos

		inline val left get() = pos.left
		inline val right get() = pos.right
		inline val top get() = pos.top
		inline val bottom get() = pos.bottom
		inline val topLeft get() = pos.topLeft
		inline val topRight get() = pos.topRight
		inline val bottomLeft get() = pos.bottomLeft
		inline val bottomRight get() = pos.bottomRight

		/**
		 * Хранит тайлы как x + y * CHUNK_SIZE
		 */
		private val tiles = arrayOfNulls<ChunkTile>(CHUNK_SIZE * CHUNK_SIZE)

		override operator fun get(x: Int, y: Int): ChunkTile? {
			if (isOutside(x, y))
				return null

			return tiles[x or (y shl CHUNK_SHIFT)]
		}

		operator fun set(x: Int, y: Int, tile: ChunkTile?) {
			if (isOutside(x, y))
				throw IndexOutOfBoundsException("Trying to set tile ${tile?.def?.materialName} at $x $y, but that is outside of chunk's range")

			changeset++
			tiles[x or (y shl CHUNK_SHIFT)] = tile
			markPhysicsDirty()
		}

		override operator fun set(x: Int, y: Int, tile: TileDefinition?): ChunkTile? {
			if (isOutside(x, y))
				throw IndexOutOfBoundsException("Trying to set tile ${tile?.materialName} at $x $y, but that is outside of chunk's range")

			val chunkTile = if (tile != null) ChunkTile(this, tile) else null
			this[x, y] = chunkTile
			changeset++
			markPhysicsDirty()
			return chunkTile
		}

		override fun randomLongFor(x: Int, y: Int): Long {
			return super.randomLongFor(x, y) xor world.seed
		}
	}

	val foreground = TileLayer()
	val background = TileLayer()
	protected val entities = HashSet<Entity>()
	val entitiesAccess = Collections.unmodifiableSet(entities)

	protected abstract fun onEntityAdded(entity: Entity)
	protected abstract fun onEntityTransferedToThis(entity: Entity, otherChunk: This)
	protected abstract fun onEntityTransferedFromThis(entity: Entity, otherChunk: This)
	protected abstract fun onEntityRemoved(entity: Entity)

	fun addEntity(entity: Entity) {
		if (!entities.add(entity)) {
			throw IllegalArgumentException("Already having having entity $entity")
		}

		onEntityAdded(entity)
	}

	fun transferEntity(entity: Entity, otherChunk: Chunk<*, *>) {
		if (otherChunk == this)
			throw IllegalArgumentException("what?")

		if (this::class.java != otherChunk::class.java) {
			throw IllegalArgumentException("Incompatible types: $this !is $otherChunk")
		}

		if (!entities.add(entity)) {
			throw IllegalArgumentException("Already containing $entity")
		}

		onEntityTransferedToThis(entity, otherChunk as This)
		otherChunk.onEntityTransferedFromThis(entity, this as This)

		if (!otherChunk.entities.remove(entity)) {
			throw IllegalStateException("Unable to remove $entity from $otherChunk after transfer")
		}
	}

	fun removeEntity(entity: Entity) {
		if (!entities.remove(entity)) {
			throw IllegalArgumentException("Already not having entity $entity")
		}

		onEntityRemoved(entity)
	}

	override fun toString(): String {
		return "Chunk(pos=$pos, entityCount=${entities.size}, world=$world)"
	}

	companion object {
		val EMPTY = object : IMutableTileChunk {
			override val pos = ChunkPos(0, 0)

			override fun get(x: Int, y: Int): ChunkTile? = null
			override fun set(x: Int, y: Int, tile: TileDefinition?): ChunkTile? = null
		}

		private val aabbBase = AABB(
			Vector2d.ZERO,
			Vector2d(CHUNK_SIZE.toDouble(), CHUNK_SIZE.toDouble()),
		)
	}
}
